@@ -1,6 +1,7 @@
 from six import iteritems, string_types
 
 from elasticsearch.helpers import scan
+from elasticsearch.exceptions import TransportError
 
 from .query import Q, EMPTY_QUERY, Bool
 from .aggs import A, AggBase
@@ -96,8 +97,8 @@ class Request(object):
         """
         Specify query params to be used when executing the search. All the
         keyword arguments will override the current values. See
-        http://elasticsearch-py.readthedocs.org/en/master/api.html#elasticsearch.Elasticsearch.search
-        for all availible parameters.
+        https://elasticsearch-py.readthedocs.io/en/master/api.html#elasticsearch.Elasticsearch.search
+        for all available parameters.
 
         Example::
 
@@ -110,7 +111,7 @@ class Request(object):
 
     def index(self, *index):
         """
-        Set the index for the search. If called empty it will rmove all information.
+        Set the index for the search. If called empty it will remove all information.
 
         Example:
 
@@ -203,13 +204,14 @@ class Search(Request):
         :arg index: limit the search to index
         :arg doc_type: only query this type.
 
-        All the paramters supplied (or omitted) at creation type can be later
+        All the parameters supplied (or omitted) at creation type can be later
         overriden by methods (`using`, `index` and `doc_type` respectively).
         """
         super(Search, self).__init__(**kwargs)
 
         self.aggs = AggsProxy(self)
         self._sort = []
+        self._source = None
         self._fields = None
         self._partial_fields = {}
         self._highlight = {}
@@ -265,7 +267,7 @@ class Search(Request):
     @classmethod
     def from_dict(cls, d):
         """
-        Construct a `Search` instance from a raw dict containing the search
+        Construct a new `Search` instance from a raw dict containing the search
         body. Useful when migrating from raw dictionaries.
 
         Example::
@@ -294,7 +296,8 @@ class Search(Request):
 
         s._response_class = self._response_class
         s._sort = self._sort[:]
-        s._fields = self._fields[:] if self._fields is not None else None
+        s._source = self._source.copy() if self._source else None
+        s._fields = self._fields[:] if self._fields else None
         s._partial_fields = self._partial_fields.copy()
         s._highlight = self._highlight.copy()
         s._highlight_opts = self._highlight_opts.copy()
@@ -335,6 +338,8 @@ class Search(Request):
             }
         if 'sort' in d:
             self._sort = d.pop('sort')
+        if '_source' in d:
+            self._source = d.pop('_source')
         if 'fields' in d:
             self._fields = d.pop('fields')
         if 'partial_fields' in d:
@@ -352,6 +357,7 @@ class Search(Request):
         if 'script_fields' in d:
             self._script_fields = d.pop('script_fields')
         self._extra = d
+        return self
 
     def script_fields(self, **kwargs):
         """
@@ -378,6 +384,44 @@ class Search(Request):
         s._script_fields.update(kwargs)
         return s
 
+    def source(self, **kwargs):
+        """
+        Selectively control how the _source field is returned.
+
+        :arg source: wildcard string, array of wildcards, or dictionary of includes and excludes
+
+        If ``source`` is None, the entire document will be returned for
+        each hit.  If source is a dictionary with keys of 'include' and/or
+        'exclude' the fields will be either included or excluded appropriately.
+
+        Calling this multiple times with the same named parameter will override the
+        previous values with the new ones.
+
+        Example::
+
+            s = Search()
+            s = s.source(include=['obj1.*'], exclude=["*.description"])
+
+            s = Search()
+            s = s.source(include=['obj1.*']).source(exclude=["*.description"])
+
+        """
+        s = self._clone()
+
+        if s._source is None:
+            s._source = {}
+
+        for key, value in kwargs.items():
+            if value is None:
+                try:
+                    del s._source[key]
+                except KeyError:
+                    pass
+            else:
+                s._source[key] = value
+
+        return s
+
     def fields(self, fields=None):
         """
         Selectively load specific stored fields for each document.
@@ -460,7 +504,7 @@ class Search(Request):
 
     def highlight(self, *fields, **kwargs):
         """
-        Request highliting of some fields. All keyword arguments passed in will be
+        Request highlighting of some fields. All keyword arguments passed in will be
         used as parameters. Example::
 
             Search().highlight('title', 'body', fragment_size=50)
@@ -492,7 +536,7 @@ class Search(Request):
         All keyword arguments will be added to the suggestions body. For example::
 
             s = Search()
-            s = s.suggest('suggestion-1', 'Elasticserach', term={'field': 'body'})
+            s = s.suggest('suggestion-1', 'Elasticsearch', term={'field': 'body'})
         """
         s = self._clone()
         s._suggest[name] = {'text': text}
@@ -524,6 +568,9 @@ class Search(Request):
 
             d.update(self._extra)
 
+            if self._source:
+                d['_source'] = self._source
+
             if self._fields is not None:
                 d['fields'] = self._fields
 
@@ -558,7 +605,8 @@ class Search(Request):
         return es.count(
             index=self._index,
             doc_type=self._doc_type,
-            body=d
+            body=d,
+            **self._params
         )['count']
 
     def execute(self, ignore_cache=False):
@@ -603,7 +651,7 @@ class Search(Request):
 
         Use ``params`` method to specify any additional arguments you with to
         pass to the underlying ``scan`` helper from ``elasticsearch-py`` -
-        http://elasticsearch-py.readthedocs.org/en/master/helpers.html#elasticsearch.helpers.scan
+        https://elasticsearch-py.readthedocs.io/en/master/helpers.html#elasticsearch.helpers.scan
 
         """
         es = connections.get_connection(self._using)
@@ -654,7 +702,7 @@ class MultiSearch(Request):
 
         return out
 
-    def execute(self, ignore_cache=False):
+    def execute(self, ignore_cache=False, raise_on_error=True):
         if ignore_cache or not hasattr(self, '_response'):
             es = connections.get_connection(self._using)
 
@@ -667,8 +715,14 @@ class MultiSearch(Request):
 
             out = []
             for s, r in zip(self._searches, responses['responses']):
-                r = Response(r, callbacks=s._doc_type_map)
-                r.search = s
+                if r.get('error', False):
+                    print(r)
+                    if raise_on_error:
+                        raise TransportError('N/A', r['error']['type'], r['error'])
+                    r = None
+                else:
+                    r = Response(r, callbacks=s._doc_type_map)
+                    r.search = s
                 out.append(r)
 
             self._response = out
